#!/usr/bin/env python

# Copyright © 2024 Andreas Krüger, DJ3EI.
# 
# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
import logging
import random
import re
import subprocess
from enum import Enum
from functools import reduce
from pathlib import Path
from typing import List, Tuple, Union, Dict, Iterable, Optional, Literal, cast

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
LOGGER = logging.getLogger("main")

DEPENDENCY_VERSION = Union[Tuple[int], Tuple[int, int], Tuple[int, int, int]]


class VersionState(Enum):
    GOOD = 0  # At least one combination with versions of other dependencies and this version succeeded.
    CANDIDATE = 1  # We don't know anything.
    PIP_FAILED = 2  # At least in one context, pip could not install this version
    TEST_FAILED = 3  # A test with this version the culprit for test failure.


class Dependency:
    NUM_OF_CONSIDERATIONS = 5
    INITIAL_TOTAL_PROBABILITY = 1 / 10
    INITIAL_PROBABILITY = 1 - (1 - INITIAL_TOTAL_PROBABILITY) ** (1 / NUM_OF_CONSIDERATIONS)

    LOGGER.info("INITIAL_PROBABILITY %f should result in %f",
                INITIAL_PROBABILITY,
                1 - (1 - INITIAL_TOTAL_PROBABILITY)**5)

    def __init__(self, name: str, version_names: Iterable[str]):
        split_version_re = re.compile(r"^(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)")
        self.name = name
        production_versions: List[DEPENDENCY_VERSION] = []  # sorted youngest first.
        for version_name in version_names:
            mo = split_version_re.fullmatch(version_name)
            if mo is not None:
                if mo.group("patch") is not None:
                    production_versions.append(
                        (int(mo.group("major")), int(mo.group("minor")), int(mo.group("patch")),))
                elif mo.group("minor") is not None:
                    production_versions.append((int(mo.group("major")), int(mo.group("minor")),))
                else:
                    production_versions.append((int(mo.group("major")),))
            else:
                LOGGER.debug("%s non-production version %s ignored",
                             name, version_name.replace("\n", r"\n"))
        self.production_versions = sorted(production_versions, reverse=True)
        if 0 == len(self.production_versions):
            raise ValueError(f"No versions found for dependency {name}")
        LOGGER.info("%s: %i production versions found", name, len(self.production_versions))
        self.version2state: Dict[DEPENDENCY_VERSION, Union[
            Tuple[Literal[VersionState.GOOD], None],
            Tuple[Literal[VersionState.CANDIDATE], float],  # probability this might get chosen
            Tuple[Literal[VersionState.TEST_FAILED], None],
            Tuple[Literal[VersionState.PIP_FAILED], None]
        ]] = {}
        self.oldest_good_version = self.production_versions[0]
        self.oldest_good_version_index = 0

        for version in self.production_versions:
            self.version2state[version] = (VersionState.CANDIDATE, Dependency.INITIAL_PROBABILITY)

        self.candidates_under_consideration: List[DEPENDENCY_VERSION] = []
        self.register_good_version(self.oldest_good_version, True)

    def version_to_pip_string(self, dep_version: DEPENDENCY_VERSION, lenient: bool = True, quoted: bool = False) -> str:
        version_string = ".".join((str(d) for d in dep_version))
        if lenient:
            result = f"{self.name}<={version_string}"
        else:
            result = f"{self.name}=={version_string}"
        if quoted:
            return f'"{result}"'
        else:
            return result

    def _boost_or_dampen(self, boost: bool = True) -> None:
        # Boost those we consider a bit:
        for candidate in self.candidates_under_consideration:
            sp = self.version2state[candidate]
            if sp[0] == VersionState.CANDIDATE:
                probability: float = sp[1]
                if boost:
                    self.version2state[candidate] = (VersionState.CANDIDATE, probability ** 0.7)
                else:
                    self.version2state[candidate] = (VersionState.CANDIDATE, probability / 5.0)
            else:
                LOGGER.warning("Unexpected non-candidacy state %i for %s",
                               sp[0],
                               self.version_to_pip_string(candidate, False))

    def _consider(self, dep_version: DEPENDENCY_VERSION) -> bool:
        if dep_version in self.candidates_under_consideration:
            LOGGER.info("Not considering anew: %s - already candidate.",
                        self.version_to_pip_string(dep_version, False))
            return False
        elif Dependency.NUM_OF_CONSIDERATIONS <= len(self.candidates_under_consideration):
            raise RuntimeError("Trying to add more considerations on top of those we already have.")
        elif self.version2state[dep_version][0] != VersionState.CANDIDATE:
            LOGGER.info("Not considering anew: %s - no candidate status.",
                        self.version_to_pip_string(dep_version, False))
            return False
        else:
            LOGGER.info("Considering anew: %s", self.version_to_pip_string(dep_version, False))
            self.candidates_under_consideration.append(dep_version)
            return True

    def register_pip_failure(self, dep_version: DEPENDENCY_VERSION) -> None:
        self.candidates_under_consideration = [v for v in self.candidates_under_consideration if v != dep_version]
        self.version2state[dep_version] = (VersionState.PIP_FAILED, None)
        # Dampen a bit:
        self._boost_or_dampen(False)

    def register_test_failure_and_its_this_versions_fault(self, dep_version: DEPENDENCY_VERSION) -> None:
        self.candidates_under_consideration = [v for v in self.candidates_under_consideration if v != dep_version]
        self.version2state[dep_version] = (VersionState.TEST_FAILED, None)
        # Dampen a bit:
        self._boost_or_dampen(False)

    def register_test_failure_caused_by_who_knows_what(self, dep_version: DEPENDENCY_VERSION) -> None:
        old_vs = self.version2state[dep_version]
        if old_vs[0] == VersionState.CANDIDATE:
            self._boost_or_dampen(False)
        else:
            LOGGER.warning("Ignoring failure for version %s with state %s",
                           self.version_to_pip_string(dep_version, False),
                           old_vs[0])

    def register_good_version(self, dep_version: DEPENDENCY_VERSION, init: bool = False) -> None:
        if self.oldest_good_version != dep_version or init:
            LOGGER.info("New good version %s", self.version_to_pip_string(dep_version, False))

            if self.oldest_good_version < dep_version:
                LOGGER.warning(
                    "Confusion: previously had good %s, now newer %s",
                    self.version_to_pip_string(self.oldest_good_version, False),
                    self.version_to_pip_string(dep_version, False))
                return

            self.oldest_good_version = dep_version
            while self.production_versions[self.oldest_good_version_index] != dep_version:
                self.oldest_good_version_index += 1

            previous_state, previous_probability = self.version2state[dep_version]
            if previous_state != VersionState.CANDIDATE:
                previous_probability = Dependency.INITIAL_PROBABILITY
            self.version2state[dep_version] = (VersionState.GOOD, None)
            if previous_probability is None:
                raise RuntimeError("This cannot happen: previous probability is None, state %i", previous_state)

            # No need to further investigate younger stuff:
            self.candidates_under_consideration = [v for v in self.candidates_under_consideration if v < dep_version]

            num_of_versions = len(self.production_versions)

            # First attempt: Add the next older version:
            if len(self.candidates_under_consideration) < Dependency.NUM_OF_CONSIDERATIONS:
                search_index_1 = self.oldest_good_version_index + 1
                if search_index_1 < num_of_versions:
                    maybe_consider_1 = self.production_versions[search_index_1]
                    if self._consider(maybe_consider_1):
                        self.version2state[maybe_consider_1] = (VersionState.CANDIDATE, previous_probability)

            # Second attempt: Add the next oldest version that differs in major and minor:
            if 2 <= len(dep_version) and len(self.candidates_under_consideration) < Dependency.NUM_OF_CONSIDERATIONS:
                major = dep_version[0]
                minor = dep_version[1]
                for search_index_2 in range(self.oldest_good_version_index, num_of_versions):
                    maybe_consider_2 = self.production_versions[search_index_2]
                    if maybe_consider_2[0] != major or (2 <= len(maybe_consider_2) and maybe_consider_2[1] != minor):
                        if self._consider(maybe_consider_2):
                            self.version2state[maybe_consider_2] = (VersionState.CANDIDATE, previous_probability)
                            break

            # Finally, add as they come:
            search_index_3 = self.oldest_good_version_index + 1
            while len(self.candidates_under_consideration) < Dependency.NUM_OF_CONSIDERATIONS and \
                    search_index_3 < num_of_versions:
                maybe_consider_3 = self.production_versions[search_index_3]
                if self._consider(maybe_consider_3):
                    self.version2state[maybe_consider_3] = (VersionState.CANDIDATE, previous_probability)
                search_index_3 += 1

            # Boost a bit:
            if not init:
                self._boost_or_dampen(True)

    def pick_version_to_test(self) -> DEPENDENCY_VERSION:
        probability_of_using_good_value = 1.0
        for candidate in self.candidates_under_consideration:
            vs = self.version2state[candidate]
            if vs[0] == VersionState.CANDIDATE:
                probability = vs[1]
                probability_of_using_good_value *= (1 - probability)
        LOGGER.debug("Probability to try %s is %f", self.name, 1 - probability_of_using_good_value)
        if random.random() <= probability_of_using_good_value:
            result = self.oldest_good_version
            LOGGER.debug("Trying good value %s", self.version_to_pip_string(result))
        else:
            result = self.pick_version_to_test_thats_not_known_good()
            LOGGER.info("Trying value %s", self.version_to_pip_string(result))
        return result

    def pick_version_to_test_thats_not_known_good(self) -> DEPENDENCY_VERSION:
        if 0 == len(self.candidates_under_consideration):
            raise RuntimeError("No choice that's not known good left.")
        sum_of_probabilites = 0.0
        for candidate in self.candidates_under_consideration:
            probability = self.version2state[candidate][1]
            sum_of_probabilites += cast(float, probability)
        r = random.uniform(0, sum_of_probabilites)
        running_sum = 0.0
        for candidate in self.candidates_under_consideration:
            running_sum += cast(float, self.version2state[candidate][1])
            if r <= running_sum:
                return candidate
        LOGGER.warning("pick_version_to_test, should not happen: random %f running_sum %f", r, running_sum)
        return self.candidates_under_consideration[-1]

    def has_version_to_test(self) -> bool:
        return 0 < len(self.candidates_under_consideration)


def read_from_json(json_file: Union[str, Path]) -> List[Dependency]:
    result: List[Dependency] = []
    with open(json_file) as json_stream:
        dep_name2version_time: Dict[str, List[Tuple[str, str]]] = json.load(json_stream)
        for dep, vts in dep_name2version_time.items():
            result.append(Dependency(dep, (vt[0] for vt in vts)))
    return result


STATUS_AND_CANDIDATE = Union[
    Tuple[Literal["ONE_CHANGED"], Tuple[DEPENDENCY_VERSION, ...], DEPENDENCY_VERSION, Dependency],
    Tuple[Literal["SEVERAL_CHANGED"], Tuple[DEPENDENCY_VERSION, ...]],
    Tuple[Literal["OLDEST_GOOD_VERSIONS"], Tuple[DEPENDENCY_VERSION, ...]]
]


class Search:
    def __init__(self, dependencies: List[Dependency],
                 activator_script: Path = Path(__file__).parent.parent.parent / "venv-find" / "bin" / "activate",
                 log_file_name: Path = Path(__file__).parent.parent.parent / "oldest_req_search_2.log"):
        self.dependencies = tuple(dependencies)
        self.results: Dict[Tuple[DEPENDENCY_VERSION, ...], VersionState] = {}
        self.activator_script = activator_script
        self.log_file_name = log_file_name

    def all_good_versions(self) -> Tuple[DEPENDENCY_VERSION, ...]:
        return tuple(d.oldest_good_version for d in self.dependencies)

    def find_candidate_versions_and_maybe_book(self) -> STATUS_AND_CANDIDATE:
        all_good = self.all_good_versions()

        # This is what we may test:
        candidate = tuple(d.pick_version_to_test() for d in self.dependencies)

        # Pick all dependencies for which the version was changed, plus versions.
        def pick_changed(
                in_acc: List[Tuple[Dependency, DEPENDENCY_VERSION]],
                in_tuple: Tuple[DEPENDENCY_VERSION, DEPENDENCY_VERSION, Dependency]) -> \
                List[Tuple[Dependency, DEPENDENCY_VERSION]]:
            if in_tuple[0] != in_tuple[1]:
                in_acc.append((in_tuple[2], in_tuple[1]))
            return in_acc

        changed_l: List[Tuple[Dependency, DEPENDENCY_VERSION]] = []
        changed_l = reduce(pick_changed, zip(all_good, candidate, self.dependencies), changed_l)

        count_changed = len(changed_l)

        if 2 <= count_changed:
            LOGGER.info("Trying to change %i of %i versions at once, namely %s",
                        count_changed,
                        len(all_good),
                        " ".join((t[0].version_to_pip_string(t[1]) for t in changed_l))
                        )
            if candidate in self.results:
                return self.find_candidate_versions_and_maybe_book()
            else:
                return "SEVERAL_CHANGED", candidate
        elif 1 == count_changed:
            changed_d = changed_l[0][0]
            changed_v = changed_l[0][1]
            if candidate in self.results:
                previous_result = self.results[candidate]
                if VersionState.GOOD == previous_result:
                    raise RuntimeError(
                        "Turns out there's only one change "
                        f"{changed_d.version_to_pip_string(changed_v)} "
                        "and we tried it before and succeeded. This cannot happen.")
                elif VersionState.PIP_FAILED == previous_result or VersionState.TEST_FAILED == previous_result:
                    # We had a failure with this previously.
                    # In the meantime, all other suspicious versions were cleared,
                    # we now know who's been at fault:
                    changed_d.register_test_failure_and_its_this_versions_fault(changed_v)
                    # So we can choose better than to re-try an old test:
                    return self.find_candidate_versions_and_maybe_book()
                else:
                    raise RuntimeError("I didn't think anything else was possible, but found {previous_result}.")
            else:
                LOGGER.info("Turns out we are trying to change only %s",
                            changed_d.version_to_pip_string(changed_v))
                if candidate in self.results:
                    return self.find_candidate_versions_and_maybe_book()
                else:
                    return "ONE_CHANGED", candidate, changed_v, changed_d
        elif 0 == count_changed:
            # Delegated random version picking has not yielded anything we'd use.
            # So let us chose one and change only that.

            # Which dependencies still have choices left?
            changeable = [d for d in self.dependencies if d.has_version_to_test()]
            length_of_changeable = len(changeable)

            if 0 < length_of_changeable:
                LOGGER.info("Presently, %i versions can be changed", length_of_changeable)
                change_me = random.choice(changeable)
                changed_version = change_me.pick_version_to_test_thats_not_known_good()
                LOGGER.info("Trying to change only %s",
                            change_me.version_to_pip_string(changed_version, False))
                candidate_with_one_change = \
                    tuple((d.oldest_good_version if d != change_me else changed_version for d in self.dependencies))
                if candidate_with_one_change in self.results:
                    # We tried that before, and know what to expect:
                    if VersionState.GOOD == self.results[candidate_with_one_change]:
                        raise RuntimeError("We know that this is good, so why does it pop up again? "
                                           f"{change_me.version_to_pip_string(changed_version)})")
                    elif VersionState.CANDIDATE == self.results[candidate_with_one_change]:
                        raise RuntimeError("This is utterly confusing. This value was never set.")
                    else:
                        change_me.register_test_failure_and_its_this_versions_fault(changed_version)
                        return self.find_candidate_versions_and_maybe_book()
                else:
                    return "ONE_CHANGED", candidate_with_one_change, changed_version, change_me
            else:
                LOGGER.info(" No versions can be changed, done with the algorithm.")
                return "OLDEST_GOOD_VERSIONS", all_good
        else:
            raise RuntimeError("This was never supposed to happen.")

    def log(self, pip_strings: Iterable[str], what: str) -> None:
        message = f"{what} {' '.join(pip_strings)}"
        with open(self.log_file_name, "a") as logf:
            logf.write(f"{message}\n")
            LOGGER.info("%s", message)

    def grab_real_versions(self) -> Tuple[DEPENDENCY_VERSION, ...]:
        pip_freeze_result = subprocess.run(
            f". {self.activator_script.absolute()} && exec pip freeze --all",
            shell=True,
            check=True,
            text=True, stdout=subprocess.PIPE,
            timeout=200
        )
        pip_output = pip_freeze_result.stdout
        linesplit = re.compile(
            r"(?P<dep_name>[a-z0-9\-_.]+)==(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)?\s*", re.IGNORECASE)
        dep_name2version: Dict[str, DEPENDENCY_VERSION] = {}
        for line in pip_output.splitlines():
            mo = linesplit.fullmatch(line)
            if mo is not None:
                dependency_name = mo.group("dep_name").lower()
                major = mo.group("major")
                minor = mo.group("minor")
                patch = mo.group("patch")
                if patch is not None:
                    dep_name2version[dependency_name] = (int(major), int(minor), int(patch))
                elif minor is not None:
                    dep_name2version[dependency_name] = (int(major), int(minor))
                else:
                    dep_name2version[dependency_name] = (int(major), )
            else:
                LOGGER.warning("Cannot parse pip freeze line %s", line.strip())

        return tuple((dep_name2version[d.name.lower()] for d in self.dependencies))

    def run(self, command: str) -> bool:
        """Try to run a shell command, returning True if it succeeded."""
        try:
            LOGGER.info("Running %s", command)
            result = subprocess.run(command, shell=True, timeout=900)
            return 0 == result.returncode
        except subprocess.TimeoutExpired:
            LOGGER.warning("Process timeout: %s", command, )
            return False

    def test_and_book(
        self,
        sc: STATUS_AND_CANDIDATE
    ) -> bool:
        """Returns whether to continue with the quest (not whether sc tested successfully)."""
        versions = sc[1]
        pip_strings = [d.version_to_pip_string(v, quoted=True) for d, v in zip(self.dependencies, versions)]
        if not self.run(f". {self.activator_script.absolute()} && "
                        f"exec pip install -U  --upgrade-strategy eager {' '.join(pip_strings)}"):
            self.results[versions] = VersionState.PIP_FAILED
            # Try whether there is one particular dependency that cannot be installed:
            found_problem = False
            for d, v in zip(self.dependencies, versions):
                pip_string = d.version_to_pip_string(v, quoted=True)
                if not self.run(f". {self.activator_script.absolute()} && exec pip install -U --upgrade-strategy eager {pip_string}"):
                    d.register_pip_failure(v)
                    self.log((pip_string,), "PIP_FAILURE")
                    found_problem = True
            if not found_problem:
                if sc[0] == "ONE_CHANGED":
                    v, d = sc[2], sc[3]
                    d.register_test_failure_and_its_this_versions_fault(v)
                    pip_string = d.version_to_pip_string(v)
                    LOGGER.info("Not a team player version: %s.", pip_string)
                    self.log(pip_string, "ONE_PIP_DOESNTFIT_FAILURE")
                else:
                    self.log(pip_strings, "COMBINDED_PIP_FAILURE")
                    for d, v in zip(self.dependencies, versions):
                        d.register_test_failure_caused_by_who_knows_what(v)
            return True

        if not self.run(f". {self.activator_script.absolute()} && exec pip check"):
            self.results[versions] = VersionState.PIP_FAILED
            if sc[0] == "ONE_CHANGED":
                v, d = sc[2], sc[3]
                d.register_test_failure_and_its_this_versions_fault(v)
                self.log((d.version_to_pip_string(v),), "ONE_PIP_CHECK_FAILURE")
            else:
                self.log(pip_strings, "PIP_CHECK_FAILURE")
                for d, v in zip(self.dependencies, versions):
                    d.register_test_failure_caused_by_who_knows_what(v)
            return True

        # Do a pip --freeze and grab what pip actually presents us with:
        real_versions = self.grab_real_versions()
        real_pip_strings = [d.version_to_pip_string(v, lenient=False, quoted=False)
                            for d, v in zip(self.dependencies, real_versions)]

        # We can now assume we have the dependencies installed as wanted.
        if self.run(f". {self.activator_script.absolute()} && exec pytest -p no:cov"):
            self.results[versions] = VersionState.GOOD
            self.results[real_versions] = VersionState.GOOD
            self.log(real_pip_strings, "TEST_PASSED")
            if sc[0] == "ONE_CHANGED":
                versions, v, d = sc[1], sc[2], sc[3]
                LOGGER.info("One proved good: %s", d.version_to_pip_string(v))
                # Pic the real version for the one we changed:
                real_v = reduce(lambda acc, dep_n_rver: acc if dep_n_rver[0] != d else dep_n_rver[1],
                                zip(self.dependencies, real_versions),
                                cast(Optional[DEPENDENCY_VERSION], None))
                if real_v is not None:
                    d.register_good_version(real_v)
                else:
                    raise RuntimeError("Cannot happen: Did not find real version for %s", d.name)
                return True
            elif sc[0] == "SEVERAL_CHANGED":
                versions = sc[1]
                for v, rv, d in zip(versions, real_versions, self.dependencies):
                    d.register_good_version(rv)
                return True
            elif sc[0] == "OLDEST_GOOD_VERSIONS":
                self.log(real_pip_strings, "ALL_DONE")
                LOGGER.info("Done")
                return False
        else:
            self.results[versions] = VersionState.TEST_FAILED
            self.results[real_versions] = VersionState.TEST_FAILED
            self.log(real_pip_strings, "TEST_FAILED")
            if sc[0] == "ONE_CHANGED":
                versions, v, d = sc[1], sc[2], sc[3]
                # LOGGER.info("One proved good: %s", d.version_to_pip_string(v))
                d.register_test_failure_and_its_this_versions_fault(v)
                return True
            elif sc[0] == "SEVERAL_CHANGED":
                versions = sc[1]
                for v, d in zip(versions, self.dependencies):
                    # LOGGER.info("This, too, proved to be good: %s", d.version_to_pip_string(v))
                    d.register_test_failure_caused_by_who_knows_what(v)
                return True
            elif sc[0] == "OLDEST_GOOD_VERSIONS":
                self.log(real_pip_strings, "THOUGHT_WE_WERE_DONE_BUT_FACE_A_FAILURE")
                LOGGER.info("Thought we were done, but the final test failed. Shit. %s",
                            " ".join(real_pip_strings))
                raise RuntimeError("All presumably good versions together fail. - Shit!")


def driver() -> None:
    dependencies = read_from_json(Path(__file__).parent.parent.parent / "pypi_requirements_available_versions.json")

    previous_work: Tuple[Tuple[str, DEPENDENCY_VERSION], ...] = (
        ("Babel", (2, 12, 0)),
        ("Jinja2", (3, 1, 0)),
        ("Markdown", (3, 0)),
        ("Pillow", (9, 2, 0)),
        ("PyRSS2Gen", (1, 1)),
        ("Pygments", (2, 12, 0)),
        ("aiohttp", (3, 8, 6)),
        ("blinker", (1, 3)),
        ("coverage", (3, 6)),
        ("docutils", (0, 19)),
        ("doit", (0, 33, 1)),
        ("feedparser", (6, 0, 2)),
        ("flake8", (0, 1)),
        ("freezegun", (0, 2, 5)),
        ("ghp-import", (0, 2, 0)),
        ("hsluv", (5, 0, 0)),
        ("html5lib", (0, 2)),
        ("ipykernel", (6, 25, 2)),
        ("lxml", (4, 5, 2)),
        ("mako", (1, 0, 9)),
        ("micawber", (0, 2, 2)),
        ("natsort", (4, 0, 0)),
        ("notebook", (6, 0, 0)),
        ("phpserialize", (1, 3)),
        ("piexif", (1, 0, 0)),
        ("pygal", (2, 0, 11)),
        ("pyphen", (0, 8)),
        ("pytest", (4, 3, 0)),
        ("pytest-cov", (0, 6)),
        ("python-dateutil", (2, 8, 2)),
        ("requests", (2, 31, 0)),
        ("ruamel.yaml", (0, 15, 98)),
        ("setuptools", (60, 8, 0)),
        ("toml", (0, 7, 1)),
        ("typogrify", (2, 0, 2)),
        ("unidecode", (0, 4, 20)),
        ("watchdog", (0, 7, 1)))

    for dependency in dependencies:
        found_it = False
        for pw in previous_work:
            if pw[0] == dependency.name:
                found_it = True
                dependency.register_good_version(pw[1])
                break
        if not found_it:
            raise RuntimeError("Could not find previous work for %s", dependency.name)

    search = Search(dependencies)
    next_try = search.find_candidate_versions_and_maybe_book()
    while search.test_and_book(next_try):
        next_try = search.find_candidate_versions_and_maybe_book()

    print(next_try)

    for d, v in zip(dependencies, next_try[1]):
        print(d.version_to_pip_string(v, lenient=False))


if __name__ == "__main__":
    driver()
