#!/usr/bin/env python

# Copyright © 2024 Andreas Krüger, DJ3EI.
# 
# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
from datetime import datetime

import bs4
import re
import requests
from pathlib import Path
from typing import List, Tuple, Dict


def find_dep_names() -> List[str]:
    """Read our requirements files and generate a list of all names of dependencies used by our tests."""
    grab_dep_name_from_line = re.compile(r"^\s*(?P<dep_name>[a-z0-9\-.]+)[>=<]", re.IGNORECASE)
    grab_included_file_from_line = re.compile(r"^\s*-r\s+(?P<included_req_file_name>[a-z0-9\-.]+\.txt)", re.IGNORECASE)
    comment_line = re.compile(r"^\s*#")
    files_wanted = ["requirements.txt", "requirements-extras.txt", "requirements-tests.txt"]
    result: List[str] = []
    for req_file_basename in files_wanted:
        req_file_fullpath = Path(__file__).parent.parent.parent / req_file_basename
        with open(req_file_fullpath) as f:
            for line in f:
                mo_dep = grab_dep_name_from_line.match(line)
                if mo_dep is not None:
                    result.append(mo_dep.group("dep_name"))
                else:
                    mo_include = grab_included_file_from_line.match(line)
                    if mo_include is not None:
                        included_req_file_name = mo_include.group("included_req_file_name")
                        if included_req_file_name in files_wanted:
                            pass
                        else:
                            raise RuntimeError(f"Found unexpected inclusion line \"{line.strip()}\"")
                    else:
                        if 0 == len(line) or line.isspace() or comment_line.match(line) is not None:
                            pass
                        else:
                            raise RuntimeError(f"Do not understand line \"{line.strip()}\"")
    return sorted(result)


def grab_versions_from_pypi(dep_name: str) -> List[Tuple[str, str]]:
    """Go to Pypi and grab all known versions of the dependency named by dep_name.

    The return value consists of (version_name, version_date) tuples.
    """
    pypi_path = f"https://pypi.org/project/{dep_name}/"
    response = requests.get(pypi_path)
    response.raise_for_status()
    html = bs4.BeautifulSoup(response.text, "lxml")
    result: List[Tuple[str, datetime]] = []
    for release_a in html.find_all("a", class_="release__card"):
        versions = release_a.find_all("p", class_="release__version")
        if 1 == len(versions):
            version = versions[0]
            version_text = version.text.strip()
        else:
            raise RuntimeError(f"Do not understand {release_a}: "
                               "No <p class='release__version'>")
        times = release_a.find_all("time")
        if 1 == len(times):
            timestamp_raw = times[0]["datetime"]
            if timestamp_raw.endswith("0000"):
                timestamp_raw = f"{timestamp_raw[0:-2]}:00"
            timestamp = datetime.fromisoformat(timestamp_raw)
        else:
            raise RuntimeError(f"Do not understand {release_a}:"
                               "No <time>")
        result.append((version_text, timestamp))
    return result


if __name__ == "__main__":

    dep_name2versions:Dict[str, List[Tuple[str, datetime]]] = {}
    for dep_name in find_dep_names():
        # Unfortunately, our target JSON does not have timestamps, so convert back to string here:
        dep_name2versions[dep_name] =\
            [(v_t[0], v_t[1].isoformat()) for v_t in grab_versions_from_pypi(dep_name)]

    outfile_name = Path(__file__).parent.parent.parent / "pypi_requirements_available_versions.json"
    with open(outfile_name, "w") as outf:
        json.dump(dep_name2versions, outf, indent=2)
